@@ -12,11 +12,32 @@ import numpy as np
 
 
 class BuildTSFIS(object):
-    def __init__(self, datapath, nr_clus, variable_names=None, merge_threshold=1.0, **kwargs):
+    """
+        Learns a a new  Takagi-Sugeno fuzzy model.
+        
+        Args:
+            datapath: The path to the csv file containing the input data (argument 'datapath' or 'dataframe' should be specified by the user).
+            dataframe: Pandas dataframe containing the input data (argument 'datapath' or 'dataframe' should be specified by the user).
+            nr_clus: Number of clusters that should be identified in the data (default = 2).
+            process_categorical: Boolean to indicate whether categorical variables should be processed (default = False).
+            method: At this moment, only Takagi Sugeno models are supported (default = 'Takagi-Sugeno')
+            variable_names: Names of the variables, if not specified the names will be read from the first row of the csv file (default = None).
+            merge_threshold: Threshold for GRABS to drop fuzzy sets from the model. If the jaccard similarity between two sets is higher than this threshold, the fuzzy set will be dropped from the model.
+            **kwargs: Additional arguments to change settings of the fuzzy model.
+
+        Returns:
+            An object containing the fuzzy model, information about its setting (such as its antecedent and consequent parameters) and the different splits of the data.
+    """
+    def __init__(self, datapath=None, dataframe=None, nr_clus=2, variable_names=None, 
+            process_categorical=False, merge_threshold=1.0, **kwargs):
+
         self.datapath = datapath
         self.nr_clus = nr_clus
         self.variable_names = variable_names
         self._antecedent_estimator = None
+        verbose = False
+
+        if 'verbose' in kwargs.keys(): verbose = kwargs['verbose']
         
         # Check keyword-arguments and complete with default settings if necessary
         if 'model_order' not in kwargs.keys(): kwargs['model_order'] = 'first' 
@@ -26,7 +47,7 @@ class BuildTSFIS(object):
             if 'sampling_number_of_bins' not in kwargs.keys(): kwargs['sampling_number_of_bins'] = 5
             if 'sampling_histogram' not in kwargs.keys(): kwargs['sampling_histogram'] = False            
         if 'data_delimiter' not in kwargs.keys(): kwargs['data_delimiter'] = ','
-        #if 'data_split_method' not in kwargs.keys(): kwargs['data_split_method'] = 'hold-out'
+        if 'data_split_method' not in kwargs.keys(): kwargs['data_split_method'] = 'hold-out'
         if 'feature_selection' not in kwargs.keys(): kwargs['feature_selection'] = None
         if 'fs_max_iter' not in kwargs.keys(): kwargs['fs_max_iter'] = 100
         if 'cluster_method' not in kwargs.keys(): kwargs['cluster_method'] = 'fcm'
@@ -45,11 +66,16 @@ class BuildTSFIS(object):
         if 'operators' not in kwargs.keys(): kwargs['operators'] = None
         if 'global_fit' not in kwargs.keys(): kwargs['global_fit'] = False  
         if 'save_simpful_code' not in kwargs.keys(): kwargs['save_simpful_code'] = True
+        if 'verbose' not in kwargs.keys(): kwargs['verbose'] = True
+        if 'cv_randomID' not in kwargs.keys(): kwargs['cv_randomID'] = False
           
         
         # Load the data
-        dl=DataLoader(self.datapath,normalize=kwargs['normalize'], delimiter=kwargs['data_delimiter'])
-        self.variable_names=dl.variable_names        
+        if self.datapath is None:
+            dl=DataLoader(dataframe=dataframe, normalize=kwargs['normalize'], process_categorical=process_categorical, delimiter=kwargs['data_delimiter'], verbose=verbose)
+        else:
+            dl=DataLoader(self.datapath, normalize=kwargs['normalize'],  process_categorical=process_categorical, delimiter=kwargs['data_delimiter'], verbose=verbose)
+        self.variable_names=dl.variable_names
 
         # Create a DataSplitter object
         ds = DataSplitter()
@@ -62,19 +88,24 @@ class BuildTSFIS(object):
         #######
         
         if kwargs['data_split_method'] == 'hold-out' or kwargs['data_split_method'] == 'holdout':
-            print('Hold-out method selected.')
+           
+            if verbose: print(' * Hold-out method selected.')
             
             # Split the data using the hold-out method in a training (default: 75%) 
             # and test set (default: 25%).
-            self.x_train, self.y_train, self.x_test, self.y_test = ds.holdout(dl.dataX, dl.dataY)
-            
+            self.x_train, self.y_train, self.x_test, self.y_test = ds.holdout(dataX=dl.dataX, dataY=dl.dataY)
+            # Check if there are any missing variables and impute them
             if np.isnan(dl.dataX).any()== True:
-                print('Warning: Your data contains missing values that will be imputed using KNN.')
-                from sklearn.impute import KNNImputer
-    
+                try:
+                    from sklearn.impute import KNNImputer
+                except ImportError:
+                    raise Exception('pyFUME tried to impute missing values, but couldn`t find \'sklearn\'. Please pip install sklearn to proceed.')
+
+                if verbose: print('Warning: Your data contains missing values that will be imputed using KNN.')   
                 imputer = KNNImputer(n_neighbors=3, weights="uniform")
                 self.x_train=imputer.fit_transform(self.x_train)
                 self.x_test=imputer.fit_transform(self.x_test)
+
             
             if kwargs['oversampling'] == True:
                 sample = Sampler(train_x = self.x_train, train_y=self.y_train, number_of_bins =  kwargs['sampling_number_of_bins'], histogram =  kwargs['sampling_histogram'])
@@ -83,7 +114,7 @@ class BuildTSFIS(object):
             # Perform feature selection if requested
             if kwargs['feature_selection'] != None:
                 if 'feature_selection_performance_metric' not in kwargs.keys(): kwargs['feature_selection_performance_metric'] = 'MAE'
-                fs=FeatureSelector(self.x_train, self.y_train, self.nr_clus, self.variable_names, model_order= kwargs['model_order'], performance_metric = kwargs['feature_selection_performance_metric'])
+                fs = FeatureSelector(self.x_train, self.y_train, self.nr_clus, self.variable_names, model_order= kwargs['model_order'], performance_metric = kwargs['feature_selection_performance_metric'], verbose=kwargs['verbose'])
                 
                 # Keep copies of the training and test set before dropping unselected features
                 self.x_train_before_fs=self.x_train.copy()
@@ -100,10 +131,10 @@ class BuildTSFIS(object):
                 self.selected_variable_names= self.variable_names
                 
             # Cluster the training data (in input-output space) using FCM
-            cl = Clusterer(self.x_train, self.y_train, self.nr_clus)
+            cl = Clusterer(x_train=self.x_train, y_train=self.y_train, nr_clus=self.nr_clus, verbose=verbose)
             
             if kwargs['cluster_method'] == 'fcm':
-                self.cluster_centers, self.partition_matrix, _ = cl.cluster(cluster_method='fcm', fcm_m=kwargs['fcm_m'], 
+                self.cluster_centers, self.partition_matrix, _ = cl.cluster(cluster_method='fcm', fcm_m=kwargs['m'], 
                     fcm_maxiter=kwargs['fcm_maxiter'], fcm_error=kwargs['fcm_error'])
             elif kwargs['cluster_method'] == 'gk':
                 self.cluster_centers, self.partition_matrix, _ = cl.cluster(cluster_method='gk')
@@ -114,17 +145,17 @@ class BuildTSFIS(object):
                 
             
             # Estimate the membership funtions of the system (default shape: gauss)
-            self._antecedent_estimator = AntecedentEstimator(self.x_train, self.partition_matrix)
+            self._antecedent_estimator = AntecedentEstimator(x_train=self.x_train, partition_matrix=self.partition_matrix)
     
             self.antecedent_parameters = self._antecedent_estimator.determineMF(mf_shape=kwargs['mf_shape'], merge_threshold=merge_threshold)
             what_to_drop = self._antecedent_estimator._info_for_simplification
             
-            # Calculate the firing strnegths
-            fsc=FireStrengthCalculator(self.antecedent_parameters, self.nr_clus, self.variable_names, **kwargs)
-            self.firing_strengths = fsc.calculate_fire_strength(self.x_train)
+            # Calculate the firing strengths
+            fsc=FireStrengthCalculator(antecedent_parameters=self.antecedent_parameters, nr_clus=self.nr_clus, variable_names=self.variable_names,  **kwargs)
+            self.firing_strengths = fsc.calculate_fire_strength(data=self.x_train)
   
             # Estimate the parameters of the consequent
-            ce = ConsequentEstimator(self.x_train, self.y_train, self.firing_strengths)
+            ce = ConsequentEstimator(x_train=self.x_train, y_train=self.y_train, firing_strengths=self.firing_strengths)
             if kwargs['model_order'] == 'first':
                 self.consequent_parameters = ce.suglms()
             elif kwargs['model_order'] == 'zero':
@@ -141,13 +172,13 @@ class BuildTSFIS(object):
                 model_order=kwargs["model_order"],
                 operators=kwargs["operators"], 
                 save_simpful_code=kwargs['save_simpful_code'], 
-                fuzzy_sets_to_drop=what_to_drop)
+                fuzzy_sets_to_drop=what_to_drop, verbose=verbose)
     
             self.model = simpbuilder.simpfulmodel
             
         elif kwargs['data_split_method']=='cross_validation' or kwargs['data_split_method']=='k-fold_cross_validation' or kwargs['data_split_method']=='crossvalidation' or kwargs['data_split_method']=='cv':
             if 'number_of_folds' not in kwargs.keys(): kwargs['number_of_folds'] = 10
-            print('K-fold cross validation was selected. The number of folds (k) equals', kwargs['number_of_folds'])
+            if verbose: print('K-fold cross validation was selected. The number of folds (k) equals', kwargs['number_of_folds'])
             
             #Create lists with test indices for each fold.
             self.fold_indices = ds.kfold(data_length=len(dl.dataX), number_of_folds=kwargs['number_of_folds'])
@@ -161,9 +192,18 @@ class BuildTSFIS(object):
                         
             # Create folder to store developed models
             import os, datetime, pickle
-            self.folder_name= 'pyFUME run ' + datetime.datetime.now().strftime("%Y-%m-%d %H.%M.%S")
-            owd = os.getcwd()
+
+            if kwargs['cv_randomID'] == True:
+                try:
+                    import uuid
+                except ImportError:
+                    raise Exception('pyFUME tried to generate random IDs, but couldn`t find \'uuid\'. Please pip install uuid to proceed.')
+                
+                self.folder_name= 'pyFUME runID ' + str(uuid.uuid4())
+            elif kwargs['cv_randomID'] == False:
+                self.folder_name= 'pyFUME run ' + datetime.datetime.now().strftime("%Y-%m-%d %H.%M.%S")
             
+            owd = os.getcwd()
             os.mkdir(self.folder_name)
             os.chdir('./' + self.folder_name)
             
@@ -171,7 +211,7 @@ class BuildTSFIS(object):
             
             
             for fold_number in range(0, kwargs['number_of_folds']):
-                print('Training the model for fold', fold_number)
+                if verbose: print('Training the model for fold', fold_number)
                 tst_idx=self.fold_indices[fold_number]
                 tst_idx = tst_idx[~np.isnan(tst_idx)]
                 tst_idx = [int(x) for x in tst_idx]
@@ -186,9 +226,12 @@ class BuildTSFIS(object):
                 self.y_test = np.array([dl.dataY[i] for i in tst_idx]) 
                 
                 if np.isnan(dl.dataX).any()== True:
-                    print('Warning: Your data contains missing values that will be imputed using KNN.')
-                    from sklearn.impute import KNNImputer
-    
+                    try:
+                        from sklearn.impute import KNNImputer
+                    except ImportError:
+                        raise Exception('pyFUME tried to impute missing values, but couldn`t find \'sklearn\'. Please pip install sklearn to proceed.')
+
+                    if verbose: print('Warning: Your data contains missing values that will be imputed using KNN.')   
                     imputer = KNNImputer(n_neighbors=3, weights="uniform")
                     self.x_train=imputer.fit_transform(self.x_train)
                     self.x_test=imputer.fit_transform(self.x_test)
@@ -199,7 +242,7 @@ class BuildTSFIS(object):
             
                 # Perform feature selection if requested
                 if kwargs['feature_selection'] != None:
-                    fs=FeatureSelector(self.x_train, self.y_train, self.nr_clus, self.variable_names)
+                    fs=FeatureSelector(self.x_train, self.y_train, self.nr_clus, self.variable_names, verbose=kwargs['verbose'])
                     self.x_train_before_fs=self.x_train.copy()
                     self.x_test_before_fs=self.x_test.copy()
                     
@@ -214,10 +257,10 @@ class BuildTSFIS(object):
                     self.selected_variable_names= self.variable_names                    
                 
                 # Cluster the training data (in input-output space) using FCM
-                cl = Clusterer(self.x_train, self.y_train, self.nr_clus)
+                cl = Clusterer(x_train=self.x_train, y_train=self.y_train, nr_clus=self.nr_clus)
                 
                 if kwargs['cluster_method'] == 'fcm':
-                    self.cluster_centers, self.partition_matrix, _ = cl.cluster(cluster_method='fcm', fcm_m=kwargs['fcm_m'], 
+                    self.cluster_centers, self.partition_matrix, _ = cl.cluster(cluster_method='fcm', fcm_m=kwargs['m'], 
                         fcm_maxiter=kwargs['fcm_maxiter'], fcm_error=kwargs['fcm_error'])
                 elif kwargs['cluster_method'] == 'fst-pso':
                     self.cluster_centers, self.partition_matrix, _ = cl.cluster(cluster_method='fstpso', 
@@ -234,8 +277,8 @@ class BuildTSFIS(object):
                 what_to_drop = self._antecedent_estimator._info_for_simplification
         
                 # Calculate the firing strnegths
-                fsc=FireStrengthCalculator(self.antecedent_parameters, self.nr_clus, self.selected_variable_names, **kwargs)
-                self.firing_strengths = fsc.calculate_fire_strength(self.x_train)
+                fsc=FireStrengthCalculator(antecedent_parameters=self.antecedent_parameters, nr_clus=self.nr_clus, variable_names=self.selected_variable_names, **kwargs)
+                self.firing_strengths = fsc.calculate_fire_strength(data=self.x_train)
                 
                 # Estimate the parameters of the consequent
                 ce = ConsequentEstimator(self.x_train, self.y_train, self.firing_strengths)
@@ -250,7 +293,8 @@ class BuildTSFIS(object):
                     extreme_values = self._antecedent_estimator._extreme_values,
                     operators=kwargs["operators"], 
                     save_simpful_code='Fold_' + str(fold_number) +'_Simpful_code.py', 
-                    fuzzy_sets_to_drop=what_to_drop)
+                    fuzzy_sets_to_drop=what_to_drop,
+                    verbose=False)
         
                 self.model = simpbuilder.simpfulmodel
                 
@@ -264,18 +308,22 @@ class BuildTSFIS(object):
                 
             # set working directory back to where script is stored
             os.chdir(owd)
-            print(self.MAE_per_fold)
+            #print(self.MAE_per_fold)
             print('The average MAE over ' + str(kwargs['number_of_folds']) +' folds is ', str(np.mean(self.MAE_per_fold)) +' (with st. dev. ' + str(np.std(self.MAE_per_fold)) + '). \nThe best model was created in fold ' +  str(np.argmin(self.MAE_per_fold)) + ' with MAE = ' + str(np.min(self.MAE_per_fold)) + '.')
                 
         elif kwargs['data_split_method'] == 'no_split':
-            print('No test data will be split off, all data will be used for training.')
+            if verbose: print('No test data will be split off, all data will be used for training.')
             
             self.x_train = dl.dataX.copy()
             self.y_train = dl.dataY.copy()
             
             if np.isnan(dl.dataX).any()== True:
-                print('Warning: Your data contains missing values that will be imputed using KNN.')
-                from sklearn.impute import KNNImputer
+                try:
+                    from sklearn.impute import KNNImputer
+                except ImportError:
+                    raise Exception('pyFUME tried to impute missing values, but couldn`t find \'sklearn\'. Please pip install sklearn to proceed.')
+
+                if verbose: print('Warning: Your data contains missing values that will be imputed using KNN.')
     
                 imputer = KNNImputer(n_neighbors=3, weights="uniform")
                 self.x_train=imputer.fit_transform(self.x_train)
@@ -287,7 +335,7 @@ class BuildTSFIS(object):
             # Perform feature selection if requested
             if kwargs['feature_selection'] != None:
                 if 'feature_selection_performance_metric' not in kwargs.keys(): kwargs['feature_selection_performance_metric'] = 'MAE'
-                fs=FeatureSelector(self.x_train, self.y_train, self.nr_clus, self.variable_names, model_order= kwargs['model_order'], performance_metric = kwargs['feature_selection_performance_metric'])
+                fs=FeatureSelector(self.x_train, self.y_train, self.nr_clus, self.variable_names, model_order= kwargs['model_order'], performance_metric = kwargs['feature_selection_performance_metric'], verbose=kwargs['verbose'])
                 
                 # Keep copies of the training and test set before dropping unselected features
                 self.x_train_before_fs=self.x_train.copy()
@@ -302,10 +350,10 @@ class BuildTSFIS(object):
                 self.selected_variable_names= self.variable_names
                 
             # Cluster the training data (in input-output space) using FCM
-            cl = Clusterer(self.x_train, self.y_train, self.nr_clus)
+            cl = Clusterer(x_train=self.x_train, y_train=self.y_train, nr_clus=self.nr_clus)
             
             if kwargs['cluster_method'] == 'fcm':
-                self.cluster_centers, self.partition_matrix, _ = cl.cluster(cluster_method='fcm', fcm_m=kwargs['fcm_m'], 
+                self.cluster_centers, self.partition_matrix, _ = cl.cluster(cluster_method='fcm', fcm_m=kwargs['m'], 
                     fcm_maxiter=kwargs['fcm_maxiter'], fcm_error=kwargs['fcm_error'])
             elif kwargs['cluster_method'] == 'gk':
                 self.cluster_centers, self.partition_matrix, _ = cl.cluster(cluster_method='gk')
@@ -343,10 +391,11 @@ class BuildTSFIS(object):
                 model_order=kwargs["model_order"],
                 operators=kwargs["operators"], 
                 save_simpful_code=kwargs['save_simpful_code'], 
-                fuzzy_sets_to_drop=what_to_drop)
+                fuzzy_sets_to_drop=what_to_drop,
+                verbose=verbose)
     
             self.model = simpbuilder.simpfulmodel
             
         else:
-            print('Invalid data splitting method chosen. Training will be aborted.')
-                               
+            print('ERROR: invalid data splitting method chosen. Training will be aborted.')
+            exit()
